/* $Id: timelist.c,v 1.10 1998/11/24 18:34:02 marcus Exp $
******************************************************************************

   Generic timing list monitor driver

   Copyright (C) 1995 Steffen Seeger	[seeger@ggi-project.org]
   Copyright (C) 1995 Andreas Beck	[becka@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger
#define	DRIVER_NAME	"Timing list monitor driver"
#define	DRIVER_REV	"$Revision: 1.10 $"

#define	DEBUG_LEVEL	0

#ifdef DEBUG
# define NEED_MONITOR_SYMS
#endif

#include "timelist.h"

extern struct kgi_monitor monitor;
extern struct kgi_preset timing[MAX_TIMINGS]; 

static int find_timing(int xwidth, int ywidth)
{
	register int i = -1, found = 0;

	while (i < MAX_TIMINGS && ! found) {

		i++;
		found = timing[i].dclk  && timing[i].x.width == xwidth 
					&& timing[i].y.width == ywidth;
	}

	return i;
} 

static int set_timing(struct kgi_preset *mode)
{
	int i = 0, found = 0;

	if (! mode->dclk) {
		
		return -EINVAL;
	}
 
	found = (i = find_timing (mode->x.width, mode->y.width)) < MAX_TIMINGS;

	if (!found) {

		/*	No entry to be replaced, thus search for a free slot.
		*/
		i = 0;
		while (i < MAX_TIMINGS && ! (found = ! timing[i].dclk)) {
			i++;
		}
	}

	if (!found) {
		
		return -ENOMEM;
	}

	timing[i] = *mode;
	return EOK;
}

static int get_timing(struct kgi_preset *mode)
{
	int i = find_timing (mode->x.width, mode->y.width);

	if (i == MAX_TIMINGS) {
		
		return -EINVAL;

	} else {

		memcpy(mode, &(timing[i]), sizeof(timing[i]));
		return EOK; 
	}
}

static int clr_timing(struct kgi_preset *mode)
{
	int i = find_timing (mode->x.width, mode->y.width);

	if (i != MAX_TIMINGS) {
		
		memset (&(timing[i]), 0, sizeof(timing[i]));
	}
	
	return EOK;
}

/*
**	KGI module interface
*/

int kgim_monitor_init(struct kgi_card *card, struct kgi_display *dpy)
{
	int i;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV " for %s %s",
		monitor.manufact, monitor.model);
	DEBUG1("(%dx%d Pixel, %d mm x %d mm, %d-%d kHz, %d-%d Hz)",
		monitor.maxdots.x, monitor.maxdots.y,
		monitor.size.x, monitor.size.y,
		monitor.hfreq.min / 1000, monitor.hfreq.max / 1000,
		monitor.vfreq.min, monitor.vfreq.max);

	NOTICE1("modes: ");
	for (i = 0; i < MAX_TIMINGS; i++) {

		if (timing[i].dclk && 
		    (monitor.dclk.min <= timing[i].dclk) && 
		    (timing[i].dclk <= monitor.dclk.max)) {

			DEBUG1("%ix%i ", timing[i].x.width, timing[i].y.width);
		}
	}
	NOTICE1("\n");

	card->monitor = &monitor;
	return EOK;
}

void kgim_monitor_done(struct kgi_display *dpy)
{
	DEBUG1("%s %s monitor driver removed.",
		monitor.manufact, monitor.model);
}

int kgim_monitor_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd *cmd)
{
	int hfreq, vfreq, h, i = 0;
	struct kgi_mode tmp = *mode;
	static struct kgi_mode save;

	if ((i = find_timing(mode->x.width, mode->y.width)) == MAX_TIMINGS) {

		return -E(MONITOR, UNKNOWN);
	}

	switch (*cmd) {

	case CMD_PROPOSE:		/* Enter timing found */
		mode->dclk = timing[i].dclk;
		mode->x    = timing[i].x;
		mode->y    = timing[i].y;

		/* !!! Hack for the Mystique */
		if (!strcmp(dpy->card->model,"Mystique") && 
		    mode->request.graphtype == KGIGT_8BIT) {
			mode->x.total     = (mode->x.total     + 15) & ~15;
			mode->x.syncstart = (mode->x.syncstart + 15) & ~15;
			mode->x.syncend   = (mode->x.syncend   + 15) & ~15;
		}

		if (monitor.flags & FL_MON_PROPSIZE) {

			mode->request.size.x = mode->x.width * monitor.size.x
				/ monitor.maxdots.x;
			mode->request.size.y = mode->y.width * monitor.size.y
				/ monitor.maxdots.y;

		} else {

			mode->request.size = monitor.size;
		}

		DEBUG2("Propose dclk = %i Hz hfreq = %i Hz "
			"vfreq = %i Hz\n", mode->dclk,
			mode->dclk / mode->x.total,
			mode->dclk / (mode->x.total*mode->y.total));

		/*	NOTE: This depends on the kgim_check_mode()
		**	implementation!
		*/
		save = *mode;
		*cmd = CMD_LOWER;
		return EOK;

	case CMD_LOWER:
		tmp = save;
		save = *mode;
		*mode = tmp;
		*cmd  = CMD_RAISE;
		return EOK;

	case CMD_RAISE:
		/*	Take the timing whose DCLK is closer to the preset
		*/
		if (abs(save.dclk - timing[i].dclk)
			< abs(mode->dclk - timing[i].dclk)) {

			*mode = save;
		}

		memset(&save, 0, sizeof(save));
		*cmd = CMD_CHECK;
		return EOK;

     case CMD_CHECK:
		if ((mode->dclk < monitor.dclk.min) || 
		    (monitor.dclk.max < mode->dclk)) {

			DEBUG2("DCLK of %i Hz is out of bounds.", mode->dclk);
			return -E(MONITOR, UNKNOWN);
		}

		if ((monitor.maxdots.x < (int)mode->x.width) || 
		    (monitor.maxdots.y < (int)mode->y.width)) {

			DEBUG2("Resolution of %ix%i pixels is too high.",
				mode->x.width, mode->y.width);
			return -E(MONITOR, UNKNOWN);
		}

		if (!(mode->x.total && mode->y.total)) {

			return -E(MONITOR, UNKNOWN);
		}

		hfreq = mode->dclk / mode->x.total;
		vfreq = hfreq / mode->y.total;
		if ((100*abs(mode->dclk-timing[i].dclk) > 2*timing[i].dclk) ||
		    (hfreq < monitor.hfreq.min) || (monitor.hfreq.max < hfreq) ||
		    (vfreq < monitor.vfreq.min) || (monitor.vfreq.max < vfreq))
		{

			DEBUG2("timing limits violated, hfreq = %i, vfreq = %i,"
				" dclk = %i", hfreq, vfreq, mode->dclk);
			return -E(MONITOR, UNKNOWN);
		}
	
		h = (mode->request.graphtype == KGIGT_TEXT16) 
			? (mode->request.dpp.x)
			: (mode->request.dpp.x ? mode->request.dpp.x : 1);

		#define	ERR(X) (   (mode->y.X != timing[i].y.X)  \
				|| (abs(mode->x.X - timing[i].x.X) > h))
		if (   ERR(width)   || ERR(blankstart) || ERR(syncstart) 
		    || ERR(syncend) || ERR(blankend)   || ERR(total)) {

			NOTICE2("Error in timings.");
			return -E(MONITOR, UNKNOWN);
		}
		#undef ERR

		*cmd = CMD_READY;
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(MONITOR, UNKNOWN);
	}

	#undef VISX
	#undef VISY
}

void kgim_monitor_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
}

int kgim_monitor_command (struct kgi_graphic *state, kgi_uint cmd, kgi_uint iobuf)
{
  int i, result;

	switch(cmd) {
		
	case MONITOR_GETINFO:
		memcpy(&monitor, state->io_res = state->io_buf, 
		       sizeof(monitor));
		return EOK;

	case MONITOR_SETINFO:
		if (!suser()) {
			
			return -EPERM;
		}
		memcpy(&monitor, state->io_buf, sizeof(monitor));
		return EOK;

	case MONITOR_GETNUMPRESET:
		return MAX_TIMINGS;

	case MONITOR_GETFREEPRESET:
		for (i = result = 0; i < MAX_TIMINGS; i++)
			result += timing[i].dclk ? 0 : 1;
		return result;

	case MONITOR_SETPRESETMODE:
		if (!suser()) {
			
			return -EPERM;
		}
		
		return set_timing((struct kgi_preset *) state->io_buf);

	case MONITOR_GETPRESETMODE:
		return get_timing((struct kgi_preset *) state->io_res =
			state->io_buf);

	case MONITOR_CLEARPRESETENT:
		return clr_timing((struct kgi_preset *) state->io_buf);

	default:
		return -E(MONITOR, UNKNOWN);
	}
	return -E(MONITOR, UNKNOWN);
}

#ifndef NO_MSUPARAMS
struct kgi_subparam *kgim_monitor_get_params(void)
{
	return NULL;
}
#endif
